from typing import Callable, Optional

from bottleneck import (
    allnan,
    move_argmax,
    move_argmin,
    move_max,
    move_mean,
    move_median,
    move_min,
    move_rank,
    move_std,
    move_sum,
    move_var,
    nanmax,
    nanmean,
    nanmedian,
    nanmin,
    nanstd,
    nansum,
)
from numpy import (  # pylint: disable=redefined-builtin
    abs,
    asarray,
    concatenate,
    float64,
    full,
    isnan,
    log,
    maximum,
    nan,
    nancumsum,
    nanpercentile,
    sqrt,
    zeros,
)
from numpy.lib.stride_tricks import sliding_window_view
from numpy.typing import NDArray
from scipy.stats import kurtosis, skew

from tompy.vector import (
    __lr2sr,
    __sr2lr,
    vector_describe_meta,
    vector_lr_sr_metrics,
    vector_lr_sr_metrics_meta,
    vector_zeros,
)


def matrix_validate(m: NDArray[float64]) -> tuple[int, int]:
    assert m.ndim == 2
    return m.shape  # type: ignore


def matrix_validate_rolling(
    m: NDArray[float64], window: int
) -> tuple[int, int]:
    assert window >= 0
    return matrix_validate(m)


def matrix_zeros(nr: int, nc: int) -> NDArray[float64]:
    return zeros((nr, nc), dtype=float64)


def matrix_nans(nr: int, nc: int) -> NDArray[float64]:
    return full((nr, nc), nan, dtype=float64)


def matrix_describe_meta() -> dict[str, str]:
    return vector_describe_meta()


def matrix_describe(m: NDArray[float64]) -> NDArray[float64]:
    nr, nc = matrix_validate(m)
    if nr <= 0:
        return matrix_nans(len(matrix_describe_meta()), nc)
    return asarray(
        [
            matrix_nancnt(m),
            matrix_nanmean(m),
            matrix_nanstd(m),
            matrix_nanskew(m),
            matrix_nankurt(m),
            matrix_nanpctl(m, 0),
            matrix_nanpctl(m, 25),
            matrix_nanpctl(m, 50),
            matrix_nanpctl(m, 75),
            matrix_nanpctl(m, 100),
        ],
        dtype=float64,
    )


def matrix_shift(m: NDArray[float64], lag: int) -> NDArray[float64]:
    nr, nc = matrix_validate(m)
    if nr <= 0 or lag == 0:
        return m
    if lag > 0:
        if lag >= nr:
            return matrix_nans(nr, nc)
        return concatenate((matrix_nans(lag, nc), m[:-lag, :]))
    lag = -lag
    if lag >= nr:
        return matrix_nans(nr, nc)
    return concatenate((m[lag:, :], matrix_nans(lag, nc)))


def matrix_nancnt(m: NDArray[float64]) -> NDArray[float64]:
    nr, nc = matrix_validate(m)
    if nr <= 0:
        return matrix_nans(0, nc)
    return nr - nansum(isnan(m), axis=0)  # type: ignore


def matrix_nanmean(m: NDArray[float64]) -> NDArray[float64]:
    nr, nc = matrix_validate(m)
    if nr <= 0:
        return matrix_nans(0, nc)
    return nanmean(m, axis=0)  # type: ignore


def matrix_nanstd(m: NDArray[float64]) -> NDArray[float64]:
    nr, nc = matrix_validate(m)
    if nr <= 0:
        return matrix_nans(0, nc)
    return nanstd(m, axis=0, ddof=1)  # type: ignore


def matrix_nanskew(m: NDArray[float64]) -> NDArray[float64]:
    nr, nc = matrix_validate(m)
    if nr <= 0:
        return matrix_nans(0, nc)
    if allnan(m):
        return matrix_nans(1, nc)
    return skew(m, axis=0, bias=False, nan_policy="omit")  # type: ignore


def matrix_nankurt(m: NDArray[float64]) -> NDArray[float64]:
    nr, nc = matrix_validate(m)
    if nr <= 0:
        return matrix_nans(0, nc)
    if allnan(m):
        return matrix_nans(1, nc)
    return kurtosis(  # type: ignore
        m, axis=0, fisher=True, bias=False, nan_policy="omit"
    )


def matrix_nanpctl(m: NDArray[float64], q: int) -> NDArray[float64]:
    nr, nc = matrix_validate(m)
    if nr <= 0:
        return matrix_nans(0, nc)
    if q <= 0:
        return nanmin(m, axis=0)  # type: ignore
    if q >= 100:
        return nanmax(m, axis=0)  # type: ignore
    if q == 50:
        return nanmedian(m, axis=0)  # type: ignore
    if allnan(m):
        return matrix_nans(1, nc)
    return nanpercentile(m, q, axis=0)  # type: ignore


def matrix_rolling_sum(m: NDArray[float64], window: int) -> NDArray[float64]:
    nr, nc = matrix_validate_rolling(m, window)
    if window == 0 or window > nr:
        return matrix_nans(nr, nc)
    return move_sum(m, window, axis=0)  # type: ignore


def matrix_rolling_mean(m: NDArray[float64], window: int) -> NDArray[float64]:
    nr, nc = matrix_validate_rolling(m, window)
    if window == 0 or window > nr:
        return matrix_nans(nr, nc)
    return move_mean(m, window, axis=0)  # type: ignore


def matrix_rolling_std(m: NDArray[float64], window: int) -> NDArray[float64]:
    nr, nc = matrix_validate_rolling(m, window)
    if window == 0 or window > nr:
        return matrix_nans(nr, nc)
    return move_std(m, window, axis=0, ddof=1)  # type: ignore


def matrix_rolling_var(m: NDArray[float64], window: int) -> NDArray[float64]:
    nr, nc = matrix_validate_rolling(m, window)
    if window == 0 or window > nr:
        return matrix_nans(nr, nc)
    return move_var(m, window, axis=0, ddof=1)  # type: ignore


def matrix_rolling_min(m: NDArray[float64], window: int) -> NDArray[float64]:
    nr, nc = matrix_validate_rolling(m, window)
    if window == 0 or window > nr:
        return matrix_nans(nr, nc)
    return move_min(m, window, axis=0)  # type: ignore


def matrix_rolling_max(m: NDArray[float64], window: int) -> NDArray[float64]:
    nr, nc = matrix_validate_rolling(m, window)
    if window == 0 or window > nr:
        return matrix_nans(nr, nc)
    return move_max(m, window, axis=0)  # type: ignore


def matrix_rolling_argmin(
    m: NDArray[float64], window: int
) -> NDArray[float64]:
    nr, nc = matrix_validate_rolling(m, window)
    if window == 0 or window > nr:
        return matrix_nans(nr, nc)
    return move_argmin(m, window, axis=0)  # type: ignore


def matrix_rolling_argmax(
    m: NDArray[float64], window: int
) -> NDArray[float64]:
    nr, nc = matrix_validate_rolling(m, window)
    if window == 0 or window > nr:
        return matrix_nans(nr, nc)
    return move_argmax(m, window, axis=0)  # type: ignore


def matrix_rolling_median(
    m: NDArray[float64], window: int
) -> NDArray[float64]:
    nr, nc = matrix_validate_rolling(m, window)
    if window == 0 or window > nr:
        return matrix_nans(nr, nc)
    return move_median(m, window, axis=0)  # type: ignore


def matrix_rolling_rank(m: NDArray[float64], window: int) -> NDArray[float64]:
    nr, nc = matrix_validate_rolling(m, window)
    if window == 0 or window > nr:
        return matrix_nans(nr, nc)
    return move_rank(m, window, axis=0)  # type: ignore


def matrix_rolling_apply(
    m: NDArray[float64],
    window: int,
    func: Callable[[NDArray[float64]], NDArray[float64]],
    nfout: int,
) -> NDArray[float64]:
    nr, nc = matrix_validate_rolling(m, window)
    if window == 0 or window > nr:
        return matrix_nans(nr, nfout)
    swv = sliding_window_view(m, (window, nc)).reshape((-1, window, nc))
    ret = asarray([func(sw) for sw in swv], dtype=float64)
    assert ret.ndim == 2 and ret.shape[1] == nfout
    if window > 1:
        ret = concatenate((matrix_nans(window - 1, nfout), ret))
    return ret


def matrix_price2relative(price: NDArray[float64]) -> NDArray[float64]:
    nr, nc = matrix_validate(price)
    if nr <= 1 or nc <= 0:
        return matrix_nans(nr, nc)
    rel = price[1:, :] / price[:-1, :]
    return concatenate((matrix_nans(1, nc), rel))


def matrix_price2sr(price: NDArray[float64]) -> NDArray[float64]:
    nr, nc = matrix_validate(price)
    if nr <= 1 or nc <= 0:
        return matrix_nans(nr, nc)
    sr = price[1:, :] / price[:-1, :] - 1
    return concatenate((matrix_nans(1, nc), sr))


def matrix_price2lr(price: NDArray[float64]) -> NDArray[float64]:
    nr, nc = matrix_validate(price)
    if nr <= 1 or nc <= 0:
        return matrix_nans(nr, nc)
    lr = log(price[1:, :] / price[:-1, :])
    return concatenate((matrix_nans(1, nc), lr))


def matrix_price_metrics(
    price: NDArray[float64],
    ann_factor: float,
    sqrt_ann_factor: Optional[float] = None,
    mode: int = 0,
) -> NDArray[float64]:
    rel = matrix_price2relative(price)
    sr = rel - 1
    lr = log(rel)
    return matrix_lr_sr_metrics(
        lr, sr, ann_factor, sqrt_ann_factor=sqrt_ann_factor, mode=mode
    )


def matrix_sr2lr(sr: NDArray[float64]) -> NDArray[float64]:
    nr, nc = matrix_validate(sr)
    if nr <= 0:
        return matrix_nans(nr, nc)
    return __sr2lr(sr)


def matrix_sr_metrics(
    sr: NDArray[float64],
    ann_factor: float,
    sqrt_ann_factor: Optional[float] = None,
    mode: int = 0,
) -> NDArray[float64]:
    return matrix_lr_sr_metrics(
        matrix_sr2lr(sr),
        sr,
        ann_factor,
        sqrt_ann_factor=sqrt_ann_factor,
        mode=mode,
    )


def matrix_lr2sr(lr: NDArray[float64]) -> NDArray[float64]:
    nr, nc = matrix_validate(lr)
    if nr <= 0:
        return matrix_nans(nr, nc)
    return __lr2sr(lr)


def matrix_lr2crdd(
    lr: NDArray[float64],
) -> tuple[NDArray[float64], NDArray[float64]]:
    nr, nc = matrix_validate(lr)
    if nr <= 0:
        return matrix_nans(nr, nc), matrix_nans(nr, nc)
    lr0_cr = nancumsum(lr, axis=0)
    lr0_peak = maximum.accumulate(lr0_cr, axis=0)
    lr0_dd = lr0_cr - lr0_peak
    return lr0_cr, lr0_dd


def matrix_lr_metrics(
    lr: NDArray[float64],
    ann_factor: float,
    sqrt_ann_factor: Optional[float] = None,
    mode: int = 0,
) -> NDArray[float64]:
    return matrix_lr_sr_metrics(
        lr,
        matrix_lr2sr(lr),
        ann_factor,
        sqrt_ann_factor=sqrt_ann_factor,
        mode=mode,
    )


def matrix_lr_sr_metrics_meta(mode: int) -> dict[str, str]:
    return vector_lr_sr_metrics_meta(mode)


def matrix_lr_sr_metrics(
    lr: NDArray[float64],
    sr: NDArray[float64],
    ann_factor: float,
    sqrt_ann_factor: Optional[float] = None,
    mode: int = 0,
) -> NDArray[float64]:
    nr_lr, nc_lr = matrix_validate(lr)
    nr_sr, nc_sr = matrix_validate(sr)
    assert nr_lr == nr_sr and nc_lr == nc_sr
    if nr_lr <= 0:
        return matrix_nans(len(matrix_lr_sr_metrics_meta(mode)), nc_lr)
    if sqrt_ann_factor is None:
        sqrt_ann_factor = sqrt(ann_factor)
    return asarray(
        [
            vector_lr_sr_metrics(
                lr[:, i],
                sr[:, i],
                ann_factor,
                sqrt_ann_factor=sqrt_ann_factor,
                mode=mode,
            )
            for i in range(nc_lr)
        ],
        dtype=float64,
    ).T


def matrix_portfolio_price(  # pylint: disable=too-many-locals
    price: NDArray[float64], weight: NDArray[float64], fee_rate: float
) -> NDArray[float64]:
    price_shape = matrix_validate(price)
    weight_shape = matrix_validate(weight)
    assert price_shape == weight_shape
    nr, nc = weight_shape
    nav = 1
    position_quantity = vector_zeros(nc)
    cash = 1
    navs = vector_zeros(nr)
    for i in range(nr):
        pi = price[i]
        wi = weight[i]
        nav = cash + nansum(position_quantity * pi)
        if not allnan(wi):
            target_position_dollar = nav * wi
            target_position_quantity = target_position_dollar / pi
            trading_position_quantity = (
                target_position_quantity - position_quantity
            )
            trading_position_dollar = trading_position_quantity * pi
            trading_fee = nansum(abs(trading_position_dollar)) * fee_rate
            nav = (
                cash
                - nansum(trading_position_dollar)
                + nansum(target_position_dollar)
                - trading_fee
            )
            position_dollar = nav * wi
            position_quantity = position_dollar / pi
            cash = nav - nansum(position_dollar)
        navs[i] = nav
    return navs
