from typing import Callable, Optional

from bottleneck import (
    allnan,
    move_argmax,
    move_argmin,
    move_max,
    move_mean,
    move_median,
    move_min,
    move_rank,
    move_std,
    move_sum,
    move_var,
    nanmax,
    nanmean,
    nanmedian,
    nanmin,
    nanstd,
    nansum,
)
from numpy import (
    asarray,
    concatenate,
    exp,
    float64,
    full,
    isnan,
    log,
    maximum,
    nan,
    nancumsum,
    nanpercentile,
    sqrt,
    zeros,
)
from numpy.lib.stride_tricks import sliding_window_view
from numpy.typing import NDArray
from scipy.optimize import minimize
from scipy.stats import kurtosis, skew

from tompy.scalar import scalar_div, scalar_lr2sr


def __sr2lr(sr: NDArray[float64]) -> NDArray[float64]:
    return log(sr + 1)  # type: ignore


def __lr2sr(lr: NDArray[float64]) -> NDArray[float64]:
    return exp(lr) - 1  # type: ignore


def vector_validate(v: NDArray[float64]) -> int:
    assert v.ndim == 1
    return v.shape[0]


def vector_validate_rolling(v: NDArray[float64], window: int) -> int:
    assert window >= 0
    return vector_validate(v)


def vector_zeros(n: int) -> NDArray[float64]:
    return zeros(n, dtype=float64)


def vector_nans(n: int) -> NDArray[float64]:
    return full(n, nan, dtype=float64)


def vector_describe_meta() -> dict[str, str]:
    return {
        "COUNT": "Count",
        "MEAN": "Mean",
        "STD": "Standard Deviation",
        "SKEW": "Skewness",
        "KURT": "Kurtosis",
        "MIN": "Min",
        "25%": "25%",
        "50%": "50%",
        "75%": "75%",
        "MAX": "Max",
    }


def vector_describe(v: NDArray[float64]) -> NDArray[float64]:
    n = vector_validate(v)
    if n <= 0:
        return vector_nans(len(vector_describe_meta()))
    return asarray(
        [
            vector_nancnt(v),
            vector_nanmean(v),
            vector_nanstd(v),
            vector_nanskew(v),
            vector_nankurt(v),
            vector_nanpctl(v, 0),
            vector_nanpctl(v, 25),
            vector_nanpctl(v, 50),
            vector_nanpctl(v, 75),
            vector_nanpctl(v, 100),
        ],
        dtype=float64,
    )


def vector_shift(v: NDArray[float64], lag: int) -> NDArray[float64]:
    n = vector_validate(v)
    if n <= 0 or lag == 0:
        return v
    if lag > 0:
        if lag >= n:
            return vector_nans(n)
        return concatenate((vector_nans(lag), v[:-lag]))
    lag = -lag
    if lag >= n:
        return vector_nans(n)
    return concatenate((v[lag:], vector_nans(lag)))


def vector_nancnt(v: NDArray[float64]) -> float:
    n = vector_validate(v)
    if n <= 0:
        return nan
    return n - nansum(isnan(v))  # type: ignore


def vector_nanmean(v: NDArray[float64]) -> float:
    n = vector_validate(v)
    if n <= 0:
        return nan
    return nanmean(v)  # type: ignore


def vector_nanstd(v: NDArray[float64]) -> float:
    n = vector_validate(v)
    if n <= 0:
        return nan
    return nanstd(v, ddof=1)  # type: ignore


def vector_nanskew(v: NDArray[float64]) -> float:
    n = vector_validate(v)
    if n <= 0:
        return nan
    if allnan(v):
        return nan
    return skew(v, axis=None, bias=False, nan_policy="omit")  # type: ignore


def vector_nankurt(v: NDArray[float64]) -> float:
    n = vector_validate(v)
    if n <= 0:
        return nan
    if allnan(v):
        return nan
    return kurtosis(  # type: ignore
        v, axis=None, fisher=True, bias=False, nan_policy="omit"
    )


def vector_nanpctl(v: NDArray[float64], q: int) -> float:
    n = vector_validate(v)
    if n <= 0:
        return nan
    if q <= 0:
        return nanmin(v)  # type: ignore
    if q >= 100:
        return nanmax(v)  # type: ignore
    if q == 50:
        return nanmedian(v)  # type: ignore
    if allnan(v):
        return nan
    return nanpercentile(v, q)  # type: ignore


def vector_rolling_sum(v: NDArray[float64], window: int) -> NDArray[float64]:
    n = vector_validate_rolling(v, window)
    if window == 0 or window > n:
        return vector_nans(n)
    return move_sum(v, window)  # type: ignore


def vector_rolling_mean(v: NDArray[float64], window: int) -> NDArray[float64]:
    n = vector_validate_rolling(v, window)
    if window == 0 or window > n:
        return vector_nans(n)
    return move_mean(v, window)  # type: ignore


def vector_rolling_std(v: NDArray[float64], window: int) -> NDArray[float64]:
    n = vector_validate_rolling(v, window)
    if window == 0 or window > n:
        return vector_nans(n)
    return move_std(v, window, ddof=1)  # type: ignore


def vector_rolling_var(v: NDArray[float64], window: int) -> NDArray[float64]:
    n = vector_validate_rolling(v, window)
    if window == 0 or window > n:
        return vector_nans(n)
    return move_var(v, window, ddof=1)  # type: ignore


def vector_rolling_min(v: NDArray[float64], window: int) -> NDArray[float64]:
    n = vector_validate_rolling(v, window)
    if window == 0 or window > n:
        return vector_nans(n)
    return move_min(v, window)  # type: ignore


def vector_rolling_max(v: NDArray[float64], window: int) -> NDArray[float64]:
    n = vector_validate_rolling(v, window)
    if window == 0 or window > n:
        return vector_nans(n)
    return move_max(v, window)  # type: ignore


def vector_rolling_argmin(
    v: NDArray[float64], window: int
) -> NDArray[float64]:
    n = vector_validate_rolling(v, window)
    if window == 0 or window > n:
        return vector_nans(n)
    return move_argmin(v, window)  # type: ignore


def vector_rolling_argmax(
    v: NDArray[float64], window: int
) -> NDArray[float64]:
    n = vector_validate_rolling(v, window)
    if window == 0 or window > n:
        return vector_nans(n)
    return move_argmax(v, window)  # type: ignore


def vector_rolling_median(
    v: NDArray[float64], window: int
) -> NDArray[float64]:
    n = vector_validate_rolling(v, window)
    if window == 0 or window > n:
        return vector_nans(n)
    return move_median(v, window)  # type: ignore


def vector_rolling_rank(v: NDArray[float64], window: int) -> NDArray[float64]:
    n = vector_validate_rolling(v, window)
    if window == 0 or window > n:
        return vector_nans(n)
    return move_rank(v, window)  # type: ignore


def vector_rolling_apply(
    v: NDArray[float64],
    window: int,
    func: Callable[[NDArray[float64]], float],
) -> NDArray[float64]:
    n = vector_validate_rolling(v, window)
    if window == 0 or window > n:
        return vector_nans(n)
    swv = sliding_window_view(v, window)
    ret = asarray([func(sw) for sw in swv], dtype=float64)
    assert ret.ndim == 1
    if window > 1:
        ret = concatenate((vector_nans(window - 1), ret))
    return ret


def vector_price2relative(price: NDArray[float64]) -> NDArray[float64]:
    n = vector_validate(price)
    if n <= 1:
        return vector_nans(n)
    rel = price[1:] / price[:-1]
    return concatenate((vector_nans(1), rel))


def vector_price2sr(price: NDArray[float64]) -> NDArray[float64]:
    n = vector_validate(price)
    if n <= 1:
        return vector_nans(n)
    sr = price[1:] / price[:-1] - 1
    return concatenate((vector_nans(1), sr))


def vector_price2lr(price: NDArray[float64]) -> NDArray[float64]:
    n = vector_validate(price)
    if n <= 1:
        return vector_nans(n)
    lr = log(price[1:] / price[:-1])
    return concatenate((vector_nans(1), lr))


def vector_price_metrics(
    price: NDArray[float64],
    ann_factor: float,
    sqrt_ann_factor: Optional[float] = None,
    mode: int = 0,
) -> NDArray[float64]:
    rel = vector_price2relative(price)
    sr = rel - 1
    lr = log(rel)
    return vector_lr_sr_metrics(
        lr, sr, ann_factor, sqrt_ann_factor=sqrt_ann_factor, mode=mode
    )


def vector_price_terms_metrics(
    price: NDArray[float64],
    terms: list[int],
    ann_factor: float,
    sqrt_ann_factor: Optional[float] = None,
    mode: int = 0,
) -> NDArray[float64]:
    n = vector_validate(price)
    rel = vector_price2relative(price)
    sr = rel - 1
    lr = log(rel)
    if sqrt_ann_factor is None:
        sqrt_ann_factor = sqrt(ann_factor)
    nr = len(vector_lr_sr_metrics_meta(mode))

    def metrics(term: int) -> NDArray[float64]:
        if term > n:
            return vector_nans(nr)
        return vector_lr_sr_metrics(
            lr[-term:],
            sr[-term:],
            ann_factor,
            sqrt_ann_factor=sqrt_ann_factor,
            mode=mode,
        )

    return asarray([metrics(term) for term in terms], dtype=float64).T


def vector_sr2lr(sr: NDArray[float64]) -> NDArray[float64]:
    n = vector_validate(sr)
    if n <= 0:
        return vector_nans(n)
    return __sr2lr(sr)


def vector_sr_kelly_criterion(
    sr: NDArray[float64], ann_factor: float
) -> float:
    """
    Kelly Criterion with Empirical Distribution
    """
    n = vector_validate(sr)
    if n <= 0:
        return nan

    min_sr = nanmin(sr)
    max_sr = nanmax(sr)
    if max_sr <= 0 or min_sr >= 0:
        return nan

    eps = 0.000001
    lb = -1 / max_sr
    ub = -1 / min_sr
    bounds = [(lb + eps, ub - eps)]
    x0 = vector_zeros(1)
    result = minimize(
        lambda x: -nanmean(log(sr * x[0] + 1)) * ann_factor,
        x0,
        method="SLSQP",
        bounds=bounds,
    )
    if result.success:
        return result.x[0]  # type: ignore
    return nan


def vector_sr_metrics(
    sr: NDArray[float64],
    ann_factor: float,
    sqrt_ann_factor: Optional[float] = None,
    mode: int = 0,
) -> NDArray[float64]:
    return vector_lr_sr_metrics(
        vector_sr2lr(sr),
        sr,
        ann_factor,
        sqrt_ann_factor=sqrt_ann_factor,
        mode=mode,
    )


def vector_lr2sr(lr: NDArray[float64]) -> NDArray[float64]:
    n = vector_validate(lr)
    if n <= 0:
        return vector_nans(n)
    return __lr2sr(lr)


def vector_lr2crdd(
    lr: NDArray[float64],
) -> tuple[NDArray[float64], NDArray[float64]]:
    n = vector_validate(lr)
    if n <= 0:
        return vector_nans(n), vector_nans(n)
    lr0_cr = nancumsum(lr)
    lr0_peak = maximum.accumulate(lr0_cr)
    lr0_dd = lr0_cr - lr0_peak
    return lr0_cr, lr0_dd


def vector_lr_metrics(
    lr: NDArray[float64],
    ann_factor: float,
    sqrt_ann_factor: Optional[float] = None,
    mode: int = 0,
) -> NDArray[float64]:
    return vector_lr_sr_metrics(
        lr,
        vector_lr2sr(lr),
        ann_factor,
        sqrt_ann_factor=sqrt_ann_factor,
        mode=mode,
    )


def vector_lr_sr_metrics_meta(mode: int) -> dict[str, str]:
    if mode == 0:
        return {
            # SR Basic
            "SR_CR": "[SR] Cumulative Return = P(T) / P(0) - 1",
            "SR_AGR": "[SR] Annualized Growth Rate (aka. CAGR)",
            "SR_VOL": "[SR] Annualized Volatility",
            "SR_RDD": "[SR] Root Mean Square Drawdown (aka. Ulcer Index)",
            "SR_MDD": "[SR] Maximum Drawdown = min(SR_DD)",
            "SR_KELLY": "[SR] Kelly Criterion",
            # SR Risk Adjusted Return
            "SR_SHARPE": "[SR] Sharpe Ratio = SR_AGR / SR_VOL",
            "SR_MARTIN": "[SR] Martin Ratio = SR_AGR / SR_RDD",
        }
    return {
        # SR Basic
        "SR_CR": "[SR] Cumulative Return = P(T) / P(0) - 1",
        "SR_AGR": "[SR] Annualized Growth Rate (aka. CAGR)",
        "SR_VOL": "[SR] Annualized Volatility",
        "SR_RDD": "[SR] Root Mean Square Drawdown (aka. Ulcer Index)",
        "SR_MDD": "[SR] Maximum Drawdown = min(SR_DD)",
        "SR_KELLY": "[SR] Kelly Criterion",
        # SR Risk Adjusted Return
        "SR_SHARPE": "[SR] Sharpe Ratio = SR_AGR / SR_VOL",
        "SR_MARTIN": "[SR] Martin Ratio = SR_AGR / SR_RDD",
        # Basic
        "LR_CR": "[LR] Cumulative Return = log(P(T) / P(0))",
        "LR_AGR": "[LR] Annualized Growth Rate",
        "LR_VOL": "[LR] Annualized Volatility",
        "LR_RDD": "[LR] Root Mean Square Drawdown",
        "LR_MDD": "[LR] Maximum Drawdown = min(LR_DD)",
        # Risk Adjusted Return
        "LR_SHARPE": "[LR] Sharpe Ratio = LR_AGR / LR_VOL",
        "LR_MARTIN": "[LR] Martin Ratio = LR_AGR / LR_RDD",
        # Ratio
        "LR_WINS/TOTALS": "[LR] Wins/Totals Ratio = LR_P_COUNT / LR_COUNT",
        "LR_PROFIT/LOSS": "[LR] Profit/Loss Ratio = LR_P_MEAN / -LR_L_MEAN",
        # Etc
        "LR_DD": "[LR] Drawdown = LR_CR - LR_PEAK",
        # Descriptive Statistics
        "LR_COUNT": "[LR] Count",
        "LR_MEAN": "[LR] Mean",
        "LR_STD": "[LR] Standard Deviation",
        "LR_SKEW": "[LR] Skewness",
        "LR_KURT": "[LR] Kurtosis",
        "LR_MIN": "[LR] Min",
        "LR_25%": "[LR] 25%",
        "LR_50%": "[LR] 50%",
        "LR_75%": "[LR] 75%",
        "LR_MAX": "[LR] Max",
    }


def vector_lr_sr_metrics(  # pylint: disable=too-many-locals
    lr: NDArray[float64],
    sr: NDArray[float64],
    ann_factor: float,
    sqrt_ann_factor: Optional[float] = None,
    mode: int = 0,
) -> NDArray[float64]:
    n_lr = vector_validate(lr)
    n_sr = vector_validate(sr)
    assert n_lr == n_sr
    if n_lr <= 0:
        return vector_nans(len(vector_lr_sr_metrics_meta(mode)))
    if sqrt_ann_factor is None:
        sqrt_ann_factor = sqrt(ann_factor)

    lr0_cr, lr0_dd = vector_lr2crdd(lr)
    lr_cr = lr0_cr[-1]
    lr_rdd = sqrt(vector_nanmean(lr0_dd**2))
    lr_mdd = nanmin(lr0_dd)
    lr_mean = vector_nanmean(lr)
    lr_agr = lr_mean * ann_factor

    sr0_dd = __lr2sr(lr0_dd)
    sr_cr = scalar_lr2sr(lr_cr)
    sr_agr = scalar_lr2sr(lr_agr)
    sr_vol = vector_nanstd(sr) * sqrt_ann_factor
    sr_rdd = sqrt(vector_nanmean(sr0_dd**2))
    sr_mdd = scalar_lr2sr(lr_mdd)
    sr_kelly = vector_sr_kelly_criterion(sr, ann_factor)
    sr_sharpe = scalar_div(sr_agr, sr_vol)
    sr_martin = scalar_div(sr_agr, sr_rdd)
    if mode == 0:
        return asarray(
            [
                sr_cr,
                sr_agr,
                sr_vol,
                sr_rdd,
                sr_mdd,
                sr_kelly,
                sr_sharpe,
                sr_martin,
            ],
            dtype=float64,
        )

    lr_std = vector_nanstd(lr)
    lr_vol = lr_std * sqrt_ann_factor
    lr_sharpe = scalar_div(lr_agr, lr_vol)
    lr_martin = scalar_div(lr_agr, lr_rdd)
    lr_count = vector_nancnt(lr)
    lrp = lr[lr > 0]
    lrp_count = vector_nancnt(lrp)
    lrp_mean = vector_nanmean(lrp)
    lrl = lr[lr < 0]
    lrl_count = vector_nancnt(lrl)
    lrl_mean = vector_nanmean(lrl)
    return asarray(
        [
            sr_cr,
            sr_agr,
            sr_vol,
            sr_rdd,
            sr_mdd,
            sr_kelly,
            sr_sharpe,
            sr_martin,
            lr_cr,
            lr_agr,
            lr_vol,
            lr_rdd,
            lr_mdd,
            lr_sharpe,
            lr_martin,
            scalar_div(lrp_count, lrp_count + lrl_count),
            scalar_div(lrp_mean, -lrl_mean),
            lr0_dd[-1],
            lr_count,
            lr_mean,
            lr_std,
            vector_nanskew(lr),
            vector_nankurt(lr),
            vector_nanpctl(lr, 0),
            vector_nanpctl(lr, 25),
            vector_nanpctl(lr, 50),
            vector_nanpctl(lr, 75),
            vector_nanpctl(lr, 100),
        ],
        dtype=float64,
    )
